package cn.skcks.docking.gb28181.wvp.service.wvp;

import cn.hutool.core.date.DateUtil;
import cn.hutool.core.io.IoUtil;
import cn.hutool.crypto.digest.MD5;
import cn.skcks.docking.gb28181.common.json.JsonException;
import cn.skcks.docking.gb28181.common.json.JsonResponse;
import cn.skcks.docking.gb28181.common.json.JsonUtils;
import cn.skcks.docking.gb28181.wvp.config.WvpProxyConfig;
import cn.skcks.docking.gb28181.wvp.dto.device.DeviceChannel;
import cn.skcks.docking.gb28181.wvp.dto.device.GetDeviceChannelsReq;
import cn.skcks.docking.gb28181.wvp.dto.device.GetDeviceChannelsResp;
import cn.skcks.docking.gb28181.wvp.dto.login.WvpLoginReq;
import cn.skcks.docking.gb28181.wvp.dto.login.WvpLoginResp;
import cn.skcks.docking.gb28181.wvp.dto.record.QueryRecordReq;
import cn.skcks.docking.gb28181.wvp.dto.record.QueryRecordResp;
import cn.skcks.docking.gb28181.wvp.orm.mybatis.dynamic.model.WvpProxyDevice;
import cn.skcks.docking.gb28181.wvp.proxy.WvpProxyClient;
import cn.skcks.docking.gb28181.wvp.service.device.DeviceService;
import cn.skcks.docking.gb28181.wvp.service.download.DownloadService;
import com.github.rholder.retry.*;
import jakarta.servlet.AsyncContext;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.RequiredArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.text.MessageFormat;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

@Slf4j
@Service
@RequiredArgsConstructor
public class WvpService {
    private final WvpProxyClient wvpProxyClient;
    private final WvpProxyConfig wvpProxyConfig;
    private final DeviceService deviceService;
    private final DownloadService downloadService;

    /**
     * 默认重试次数
     */
    public final static int DEFAULT_RETRY_TIME = 3;
    /**
     * 默认每次重试等待时间
     */
    public final static int DEFAULT_RETRY_WAIT_TIME = 3;

    @SuppressWarnings("UnstableApiUsage")
    private static RetryListener defaultRetryListener(String name) {
        return new RetryListener() {
            @Override
            public <V> void onRetry(Attempt<V> attempt) {
                log.info("第 {} 次 执行 {} 结束", attempt.getAttemptNumber(), name);
                if (attempt.hasException()) {
                    log.info("执行 {} 异常 {}", name, attempt.getExceptionCause().getMessage());
                }
            }
        };
    }

    @SuppressWarnings("UnstableApiUsage")
    private static <T> Retryer<JsonResponse<T>> getDefaultRetryer(String name) {
        return RetryerBuilder.<JsonResponse<T>>newBuilder()
                // 异常就重试
                .retryIfException()
                .retryIfRuntimeException()
                // 重试间隔
                .withWaitStrategy(WaitStrategies.fixedWait(DEFAULT_RETRY_WAIT_TIME, TimeUnit.SECONDS))
                // 重试次数
                .withStopStrategy(StopStrategies.stopAfterAttempt(DEFAULT_RETRY_TIME))
                .retryIfResult(result -> result == null || (result.getCode() != 0 && result.getCode() != 200))
                .withRetryListener(defaultRetryListener(name))
                .build();
    }

    @SuppressWarnings("UnstableApiUsage")
    private static Retryer<JsonResponse<?>> getDefaultGenericRetryer(String name) {
        return RetryerBuilder.<JsonResponse<?>>newBuilder()
                // 异常就重试
                .retryIfException()
                .retryIfRuntimeException()
                // 重试间隔
                .withWaitStrategy(WaitStrategies.fixedWait(DEFAULT_RETRY_WAIT_TIME, TimeUnit.SECONDS))
                // 重试次数
                .withStopStrategy(StopStrategies.stopAfterAttempt(DEFAULT_RETRY_TIME))
                .retryIfResult(result -> result == null || (result.getCode() != 0 && result.getCode() != 200))
                .withRetryListener(defaultRetryListener(name))
                .build();
    }

    @SneakyThrows
    private void writeErrorToResponse(HttpServletResponse response, JsonResponse<?> json) {
        response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
        response.setCharacterEncoding(StandardCharsets.UTF_8.name());
        response.setContentType(MediaType.APPLICATION_JSON_VALUE);
        IoUtil.writeUtf8(response.getOutputStream(), false, json);
    }

    @SneakyThrows
    public void video(HttpServletRequest request, HttpServletResponse response, String deviceCode, Date startTime, Date endTime) {
        WvpProxyDevice wvpProxyDevice = deviceService.getDeviceByDeviceCode(deviceCode).orElse(null);
        if (wvpProxyDevice == null) {
            writeErrorToResponse(response, JsonResponse.error("设备不存在"));
            return;
        }
        String deviceId = wvpProxyDevice.getGbDeviceId();
        String channelId = wvpProxyDevice.getGbDeviceChannelId();
        log.info("设备编码 (deviceCode=>{}) 查询到的设备信息 国标id(gbDeviceId => {}), 通道(channelId => {})", deviceCode, deviceId, channelId);

        Retryer<JsonResponse<?>> genericRetryer = getDefaultGenericRetryer("调用 wvp api 查询设备历史");
        AsyncContext asyncContext = request.startAsync();
        asyncContext.setTimeout(0);
        asyncContext.start(() -> {
            HttpServletResponse asyncResponse = (HttpServletResponse) asyncContext.getResponse();
            try {
                genericRetryer.call(() -> video(asyncResponse, deviceCode, deviceId, channelId, startTime, endTime));
            } catch (RetryException e) {
                Attempt<?> failedAttempt = e.getLastFailedAttempt();
                String reason;
                if(failedAttempt.hasException()){
                    reason = MessageFormat.format("调用 wvp api 查询设备: {0} 历史失败, 已重试 {1} 次, 异常: {2}", deviceCode, e.getNumberOfFailedAttempts(), failedAttempt.getExceptionCause().getMessage());
                } else {
                    reason = MessageFormat.format("调用 wvp api 查询设备: {0} 历史失败, 已重试 {1} 次", deviceCode, e.getNumberOfFailedAttempts());
                }
                log.error(reason);
                writeErrorToResponse(asyncResponse, JsonResponse.error(reason));
            } catch (Exception e) {
                String reason = MessageFormat.format("调用 wvp api 查询设备: {0} 历史失败, 异常: {2}", deviceCode, e.getMessage());
                writeErrorToResponse(asyncResponse, JsonResponse.error(reason));
            } finally {
                asyncContext.complete();
            }
        });
    }

    /**
     * 执行 视频查询
     * <p>返回一个 JsonResponse 类型的执行结果 <p>
     * <p>如果返回值为 null 或 code 不为 0 或 200 则视为执行失败</p>
     * <p>如果排除异常 也视为执行失败</p>
     *
     * @param response   异步响应
     * @param deviceCode 设备编码 21位
     * @param deviceId   国标设备编码 20位
     * @param channelId  通道id
     * @param startTime  开始时间
     * @param endTime    结束时间
     * @return JsonResponse 类型的执行结果 如果 为 null 或 code 不为 0 或 200 则视为执行失败
     */
    @SneakyThrows
    public JsonResponse<?> video(HttpServletResponse response, String deviceCode, String deviceId, String channelId, Date startTime, Date endTime) {
        String passwdMd5 = MD5.create().digestHex(wvpProxyConfig.getPasswd());
        WvpLoginReq loginReq = WvpLoginReq.builder()
                .username(wvpProxyConfig.getUser())
                .password(passwdMd5)
                .build();
        JsonResponse<WvpLoginResp> login = wvpProxyClient.login(loginReq);
        String token = login.getData().getAccessToken();
        log.info("wvp 登录成功 token => {}", token);

        log.debug("通过 wvp 查询设备 国标id(gbDeviceId => {}) 通道信息", deviceId);
        JsonResponse<GetDeviceChannelsResp> deviceChannels = wvpProxyClient.getDeviceChannels(token, deviceId, GetDeviceChannelsReq.builder().build());
        if (deviceChannels.getData() == null || deviceChannels.getData().getTotal() == 0) {
            writeErrorToResponse(response, JsonResponse.error(MessageFormat.format("未能获取 设备: {0}, 国标id: {1}, 的通道信息", deviceCode, deviceId)));
            return JsonResponse.success(null);
        }
        List<DeviceChannel> list = deviceChannels.getData().getList();
        log.info("通过 wvp 获取到 查询设备 国标id(gbDeviceId => {}), 通道数量 => {}", deviceId, list.size());
        DeviceChannel deviceChannel = list.parallelStream().filter(item -> item.getChannelId().equalsIgnoreCase(channelId)).findFirst().orElse(null);
        if (deviceChannel == null) {
            writeErrorToResponse(response, JsonResponse.error(MessageFormat.format("未查询到 设备: {0}, 国标id: {1}, 通道: {2} 信息", deviceCode, deviceId, channelId)));
            return JsonResponse.success(null);
        }

        Retryer<JsonResponse<List<QueryRecordResp.RecordListDTO>>> queryRecordRetryer = getDefaultRetryer("调用 wvp 设备历史查询 api");
        JsonResponse<List<QueryRecordResp.RecordListDTO>> recordListResponse = queryRecordRetryer.call(() -> {
            JsonResponse<QueryRecordResp> queryRecord = wvpProxyClient.queryRecord(token, deviceId, channelId, new QueryRecordReq(DateUtil.formatDateTime(startTime), DateUtil.formatDateTime(endTime)));
            QueryRecordResp queryRecordData = queryRecord.getData();
            if (queryRecordData == null) {
                String reason = MessageFormat.format("通过 wvp 查询历史录像 失败 设备: {0}, 国标id: {1}, 通道: {2}, 错误信息: {3}", deviceCode, deviceId, channelId, queryRecord.getMsg());
                log.error(reason);
                log.error("查询历史录像 返回结果 => {}", queryRecord);
                throw new JsonException(reason);
            }
            List<QueryRecordResp.RecordListDTO> recordList = queryRecordData.getRecordList();
            if (CollectionUtils.isEmpty(recordList)) {
                String reason = MessageFormat.format("通过 wvp 查询历史录像 失败 设备: {0}, 国标id: {1}, 通道: {2}, 查询时间范围 开始时间: {3}, 结束时间: {4}, 录像数量为 0", deviceCode, deviceId, channelId, startTime, endTime);
                throw new JsonException(reason);
            }
            log.info("通过 wvp 查询到 {} 条历史录像 设备: {}, 国标id: {}, 通道: {}, 开始时间: {}, 结束时间: {}", recordList.size(), deviceCode, deviceId, channelId, startTime, endTime);
            return JsonResponse.success(recordList);
        });


        downloadService.download(response, "http://192.168.1.241:18979/download/recordTemp/0490d767d94ce20aedce57c862b6bfe9/rtp/59777645.mp4");
        return login;
    }
}
